# TODO v5+ remove this file - just a dummy to prevent import issues from v4 CLI for
# the time being


class NetworkManager:
    def __init__(self):
        pass

    def create_network(self, name: str) -> None:
        pass

    def remove_network(self, name: str) -> None:
        pass
